import click
import click_log
logger = click_log.basic_config()

import sys
from os import path, getcwd

from inspect import getsourcefile
from collections import namedtuple


current_dir = path.dirname(path.abspath(getsourcefile(lambda: 0)))
sys.path.insert(0, current_dir[: current_dir.rfind(path.sep)])


import cireqs.__version__  #pylint: disable=import-error


conf = namedtuple('Config', 'dir_path python_version timeout')


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--pythonversion', nargs=1, type=str, default='3.5.2', help='python version to use for calculating dependencies')
@click.option('--dirpath', nargs=1, help="path to directory containing requirement files, defaults to PWD")
@click.option('--timeout', nargs=1, type=int, default=10, help="how long to wait for docker commands")
@click_log.simple_verbosity_option()
@click_log.init('cireqs.cli')
def cli(ctx, dirpath, pythonversion, timeout):
    ctx.obj = conf(
        python_version=pythonversion,
        dir_path=dirpath or getcwd(),
        timeout=timeout
    )
    cireqs.set_log_level(click_log.get_level())

    splash = r"""
           o8o
           `"'
 .ooooo.  oooo  oooo d8b  .ooooo.   .ooooo oo  .oooo.o
d88' `"Y8 `888  `888""8P d88' `88b d88' `888  d88(  "8
888        888   888     888ooo888 888   888  `"Y88b.
888   .o8  888   888     888    .o 888   888  o.  )88b
`Y8bod8P' o888o d888b    `Y8bod8P' `V8bod888  8""888P'
                                         888.
                                         8P'  """
    splash = click.style(splash, fg='green') + click.style(
        "v{}".format(cireqs.__version__.__version__), fg='red') + "\n"

    if ctx.invoked_subcommand is None:
        click.echo("\n".join([splash, ctx.get_help()]))
        return


@cli.command()
@click.argument('input_requirements_filename', nargs=1, type=str, default='requirements_to_expand.txt')
@click.argument('output_requirements_filename', nargs=1, type=str, default='requirements.txt')
@click.pass_obj
def expand_requirements(conf, output_requirements_filename, input_requirements_filename):
    """Expand given requirements file by extending it using pip freeze

    args:

    input_requirements_filename: the requirements filename to expand

    output_requirements_filename: the output filename for the expanded
    requirements file
    """

    cireqs.expand_requirements(
        requirements_filename=input_requirements_filename,
        expanded_requirements_filename=output_requirements_filename,
        **conf._asdict()
    )
    logger.info("{} expanded to {} using pip freeze".format(
        input_requirements_filename, output_requirements_filename))


@cli.command()
@click.argument('input_requirements_filename', nargs=1, type=str, default='requirements.txt')
@click.pass_obj
def verify_requirements(conf, input_requirements_filename):
    """verifying that given requirements file is not missing any pins

    args:

    input_requirements_filename: requriements file to verify

    """
    cireqs.check_if_requirements_are_up_to_date(
        requirements_filename=input_requirements_filename,
        **conf._asdict())
    logger.debug("Requirements are up to date")


if __name__ == "__main__":
    cli() #pylint: disable=no-value-for-parameter



import logging
import os
import difflib
from subprocess import check_output, CalledProcessError, STDOUT
try:
    from subprocess import TimeoutExpired, DEVNULL
except ImportError:
    TimeoutExpired = None
    DEVNULL = STDOUT


logger = logging.getLogger(__name__)


def set_log_level(level):
    logger.setLevel(level)


def docker_kill_and_remove(ctr_name):
    try:
        try:
            is_running = check_output(['docker', 'top', ctr_name], stderr=DEVNULL)
        except:
            pass
        else:
            check_output(['docker', 'kill', ctr_name], stderr=STDOUT)
        try:
            has_image = check_output(['docker', 'images', '-q', ctr_name], stderr=DEVNULL)
            assert has_image
        except:
            pass
        else:
            check_output(['docker', 'rm', ctr_name], stderr=STDOUT)
    except:
        logger.error('could not stop docker container:{}'.format(ctr_name))


def docker_execute(commands, volumes=None, working_dir=None, python_version='3.5.2', timeout=10, **kwargs):
    volumes = volumes or {}
    volumes = [t for k,v in volumes.items() for t in ['-v', ':'.join([k,v])]]
    working_dir = ['-w', working_dir] if working_dir else []
    command = ' && '.join(commands)
    ctr_name = 'cireqs_container'
    docker_image = 'python:{}'.format(python_version)

    # check if has image locally:
    has_image = check_output(
        ['docker', 'images','-q', docker_image],
        stderr=DEVNULL
    )
    if not has_image:
        logger.debug('pulling docker image: {}'.format(docker_image))
        check_output(
            ['docker', 'pull', docker_image],
            stderr=STDOUT
        )

    full_command_list = [
       'docker', 'run', '--rm', '--name', ctr_name] + volumes + working_dir + [
       docker_image, 'sh', '-c', command
    ]
    logger.debug("issuing command: %s", " ".join(full_command_list))
    timeout_kwargs = {'timeout':timeout} if TimeoutExpired is not None else {}
    try:
        output = check_output(
            full_command_list,
            stderr=STDOUT,
            **timeout_kwargs
        )
        return output
    except Exception as exc:
        if isinstance(exc, CalledProcessError):
            logger.exception("Command resulted in error. " + ' '.join(full_command_list))
        elif isinstance(exc, TimeoutExpired):
            logger.warning("received timeout")
        docker_kill_and_remove(ctr_name)
        exit(-1)


def expand_requirements(dir_path, requirements_filename, expanded_requirements_filename, **kwargs):
    commands = [
        "pip install -q -r {}".format(requirements_filename),
        "pip freeze -r {} > {}".format(
            requirements_filename, expanded_requirements_filename
        )
    ]
    dir_path = os.path.normpath(dir_path) + os.sep
    working_dir = '/src'
    volumes = {
        dir_path: '/src'
    }
    docker_execute(commands, volumes, working_dir, **kwargs)
    logger.debug("Requirements expanded from {} to {}".format(
        requirements_filename, expanded_requirements_filename))


def check_if_requirements_are_up_to_date(dir_path, requirements_filename, **kwargs):
    commands = [
        "pip install -q -r {} ".format(requirements_filename),
        "pip freeze -r {} ".format(requirements_filename)
    ]
    dir_path = os.path.normpath(dir_path) + os.sep
    working_dir = '/src'
    volumes = {
        dir_path: '/src'
    }
    frozen_reqs = docker_execute(commands, volumes, working_dir, **kwargs).decode('utf-8').strip().splitlines()
    added = []
    removed = []
    with open(os.path.join(dir_path, requirements_filename), 'r') as requirements_file:
        diff_set = difflib.unified_diff(
            requirements_file.read().strip().splitlines(),
            frozen_reqs,
            fromfile='input',
            tofile='output',
            n=0,
        )
        for line in diff_set:
            for prefix in ('---', '+++', '@@', '+##', '-##'):
                if line.startswith(prefix):
                    break
            else:
                prefix, diff = line[0], line[1:]
                if prefix == '+':
                    added.append(diff)
                else:
                    removed.append(diff)
    if added or removed:
        logger.error("Requirements not up to date! added: '{}', removed: '{}'".format(
            ', '.join(added), ', '.join(removed)))
        exit(-1)



import pytest
import os.path
import sys
from cireqs.cli import cli
from click.testing import CliRunner

test_dir_path = os.path.dirname(os.path.realpath(__file__))


def test_expand_requirements():

    input_requirements_filename = 'input_requirements_1.txt'
    output_requirements_filename = 'output_requirements_1.txt'

    runner = CliRunner()
    result = runner.invoke(cli,
                           ['--dirpath', test_dir_path, 'expand_requirements',
                            input_requirements_filename,
                            output_requirements_filename])

    asserted_output = ['click-log==0.1.8\n',
                       '## The following requirements were added by pip freeze:\n',
                       'click==6.7\n']
    print(result)
    assert result.exit_code == 0
    with open(os.path.join(test_dir_path, 'output_requirements_1.txt'), 'r') as output_requirements_file:
        output_requirements_file_lines = output_requirements_file.readlines()
        assert asserted_output == output_requirements_file_lines


def test_verify_requirements_with_error():
    requirements_filename = 'input_requirements_2.txt'
    runner = CliRunner()
    result = runner.invoke(cli,
                           ['--dirpath', test_dir_path, 'verify_requirements',
                            requirements_filename])
    assert result.exit_code == -1

@pytest.mark.skipif(sys.version_info < (3,3),
                    reason="requires python3")
def test_timeout_error():
    requirements_filename = 'input_requirements_3.txt'
    runner = CliRunner()
    result = runner.invoke(cli,
                           ['--dirpath', test_dir_path, '--timeout', 0, 'verify_requirements',
                            requirements_filename])
    assert result.exit_code == -1


def test_verify_requirements():
    requirements_filename = 'input_requirements_3.txt'
    runner = CliRunner()
    result = runner.invoke(cli,
                           ['--dirpath', test_dir_path, 'verify_requirements',
                            requirements_filename])
    assert result.exit_code == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import os.path

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

readme = read_md('README.md')

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'click==6.7',
    'click-log==0.1.8',
]

test_requirements = [
    'pytest==3.2.0',
    'prospector==0.12.7'
]

extras = {
    'test': test_requirements + requirements
}

# get version
metadata = {}
version_filename = os.path.join(os.path.dirname(__file__), 'cireqs','__version__.py')
exec(open(version_filename).read(), None, metadata)

setup(
    name='cireqs',
    version=metadata['__version__'],
    description="cli tool to verify and update requirements files",
    long_description=readme + '\n\n' + history,
    author="jgv",
    author_email='jgv@trustpilot.com',
    url='https://github.com/trustpilot/python-cireqs',
    packages=[
        'cireqs',
    ],
    package_dir={'cireqs':
                 'cireqs'},
    entry_points={
        'console_scripts': [
            'cireqs=cireqs.cli:cli'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='cireqs requirements ci',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    extras_require=extras,

)
